# Merge Bump Class.  This takes a stable bump, and adds an extra
# network to force separate streams toward each other, and a third
# network to merge the two adjacent streams into 1.
#
import numpy as np
import sys
import cPickle as pickle
import pyNN.nest as sim
from nealCoverClass import NealCoverFunctions
from stateMachineClass import FSAHelperFunctions
from stableBumpClass import StableBump

class MergeBump:
    def __init__(self,NEAL,FSA):
            simName = "nest"
            spinnVersion = -1
            self.neal = NEAL
            self.fsa = FSA
            self.stableBump = StableBump(self.neal,self.fsa)
            self.hasOvercome = True

    #make a bumpAttractor object.  This should have three populations
    #and all the connectivity
    def makeMergeBump(self, numNeurons,exWindow,inWindow,exWeight,inhibWeight,
                      toOvercomeWt = 0.0013, inhibOvercomeWt = 0.4):
        self.stableBumpPop = self.stableBump.stableBump(numNeurons,
                sim.IF_cond_exp,self.fsa.CELL_PARAMS,exWindow,inWindow,
                                              exWeight,inhibWeight)
        self.inhibPop = sim.Population(numNeurons,sim.IF_cond_exp,
                                  self.fsa.CELL_PARAMS)
        self.overcomePop = sim.Population(numNeurons,sim.IF_cond_exp,
                                  self.fsa.CELL_PARAMS)

        self.connectMergeInhib (self.inhibPop,numNeurons)
        self.connectMergeAll(self.stableBumpPop,self.inhibPop,numNeurons)
        self.connectMergeOneToOne(self.stableBumpPop,self.inhibPop,numNeurons)
        self.connectMergeInhibToBump(self.inhibPop,self.stableBumpPop,
                                     numNeurons)

        #Add connectivity with the overcome network
        self.connectPopToOvercome(self.stableBumpPop,self.overcomePop,
                                  toOvercomeWt)
        self.connectOvercomeToPop(self.overcomePop,self.stableBumpPop)
        self.connectInhibOvercomeToPop(self.overcomePop,self.stableBumpPop,
                                       inhibOvercomeWt)

    #Make a bump attractor and a merge net, but no overcome
    def makeNoOvercomeBump(self, numNeurons,exWindow,inWindow,exWeight,inhibWeight,
                      toOvercomeWt = 0.0013, inhibOvercomeWt = 0.4):

        self.hasOvercome = False

        self.stableBumpPop = self.stableBump.stableBump(numNeurons,
                sim.IF_cond_exp,self.fsa.CELL_PARAMS,exWindow,inWindow,
                                              exWeight,inhibWeight)
        self.inhibPop = sim.Population(numNeurons,sim.IF_cond_exp,
                                  self.fsa.CELL_PARAMS)

        self.connectMergeInhib (self.inhibPop,numNeurons)
        self.connectMergeAll(self.stableBumpPop,self.inhibPop,numNeurons)
        self.connectMergeOneToOne(self.stableBumpPop,self.inhibPop,numNeurons)
        self.connectMergeInhibToBump(self.inhibPop,self.stableBumpPop,
                                     numNeurons)


    #These connectMerge defs are for the connectivity to merge two distant 
    #streams to two nearby streams.
    def connectMergeAll(self,fromPop,toPop,numNeurons):
        weight = 0.0006
        connector = []
        for fromNeuron in range(0,numNeurons):
            for toNeuron in range(0,numNeurons):
                    connector = connector + [(fromNeuron,toNeuron,
                                              weight, self.neal.DELAY)]

        self.neal.nealProjection(fromPop, toPop, connector, 'excitatory')

    def connectMergeInhib(self,inhibPop,numNeurons):
        weight = 0.03
        connector = []
        for neuron in range(0,numNeurons):
            if (neuron > 0):
                    connector = connector + [(neuron,neuron-1,weight, 
                                                  self.neal.DELAY)]
            if (neuron < (numNeurons -1 )):
                    connector = connector + [(neuron,neuron+1,weight, 
                                                  self.neal.DELAY)]

        #print connector
        self.neal.nealProjection(inhibPop, inhibPop, connector,'excitatory')

    def connectMergeOneToOne(self,fromPop,toPop,numNeurons):
        weight = 0.01
        connector = []
        for neuron in range(0,numNeurons):
                connector = connector + [(neuron,neuron,weight,self.neal.DELAY)]

        self.neal.nealProjection(fromPop, toPop, connector, 'excitatory')

    def connectMergeInhibToBump(self,inhibPop,toPop,numNeurons):
        inhibConnector = []
        for fromNeuron in range(0,numNeurons):
            for toNeuron in range(0,numNeurons):
                distance = abs(fromNeuron-toNeuron)
            
                base = 1.02
                if (distance > 15):
                        distanceMod = base**(distance)
                        weight = distanceMod*0.04 #paper
                        #weight = distanceMod*0.032 #for .09/.09 bump
                        #weight = distanceMod*0.029 #for .07/.07 bump
                        #print distance,distanceMod,weight
                        inhibConnector = inhibConnector + [
                                (fromNeuron,toNeuron,weight,self.neal.DELAY)]
        self.neal.nealProjection(inhibPop, toPop, inhibConnector, 'inhibitory')
    
#####
    def connectPopToOvercome(self,fromPop,toPop,weight):
        numNeurons = 100
        exciteConnector = []
        width = 10
        for fromNeuron in range(0,numNeurons):
                startTo = fromNeuron - width
                if (startTo < 0):
                        startTo = 0
                endTo = fromNeuron+width+1
                if (endTo > (numNeurons-1)):
                        endTo = numNeurons
                for toNeuron in range(startTo,endTo):
                        #weight = 0.002#0.0013
                        exciteConnector = exciteConnector + [(fromNeuron,
                                        toNeuron,weight, self.neal.DELAY)]
        self.neal.nealProjection(fromPop, toPop, exciteConnector, 'excitatory')

    def connectOvercomeToPop(self,fromPop,toPop):
        numNeurons = 100
        exciteConnector = []
        width = 3
        for fromNeuron in range(0,numNeurons):
                startTo = fromNeuron - width
                if (startTo < 0):
                        startTo = 0
                endTo = fromNeuron+width+1
                if (endTo > (numNeurons-1)):
                        endTo = numNeurons
                for toNeuron in range(startTo,endTo):
                        weight = 0.2
                        exciteConnector = exciteConnector + [(fromNeuron,
                                toNeuron,weight, self.neal.DELAY)]
        #print exciteConnector
        self.neal.nealProjection(fromPop, toPop, exciteConnector, 'excitatory')

    def connectInhibOvercomeToPop(self,fromPop,toPop, overcomeWt):
        numNeurons = 100
        inhibitConnector = []
        width = 8
        endWidth = 4
        #weight = 0.5#0.3
        for fromNeuron in range(0,numNeurons):
            startTo = fromNeuron - width
            if (startTo < 0):
                    startTo = 0
            endTo = fromNeuron - endWidth
            if (endTo < 0):
                    endTo = 0
            for toNeuron in range(startTo,endTo):
                    inhibitConnector = inhibitConnector + [
                            (fromNeuron,toNeuron,overcomeWt, self.neal.DELAY)]

            startTo = fromNeuron + endWidth+1
            if (startTo > (numNeurons-1)):
                    startTo = numNeurons
            endTo = fromNeuron + width+1
            if (endTo > (numNeurons-1)):
                    endTo = numNeurons
            for toNeuron in range(startTo,endTo):
                    inhibitConnector = inhibitConnector + [
                            (fromNeuron,toNeuron,overcomeWt, self.neal.DELAY)]

        #print inhibitConnector
        self.neal.nealProjection(fromPop, toPop, inhibitConnector, 'inhibitory')

    #allow a spike source to turn on one stableBump neurons
    def turnOnCellFromSpikeSource(self,spikeSource,cellID):
            connector = [(0,cellID,self.fsa.INPUT_WEIGHT,self.neal.DELAY)]
            self.neal.nealProjection(spikeSource,self.stableBumpPop,connector,
                                     'excitatory')

    def printResults(self, fileHeader):
        stableFileName = fileHeader + "Stable.pkl"
        self.stableBumpPop.printSpikes(stableFileName)
        stableVFileName = fileHeader + "StableV.pkl"
        self.stableBumpPop.write_data(stableVFileName,['v'])
        inhibFileName = fileHeader + "Inhib.pkl"
        self.inhibPop.printSpikes(inhibFileName)
        if (self.hasOvercome):
            overcomeFileName = fileHeader + "Overcome.pkl"
            self.overcomePop.printSpikes(overcomeFileName)

    def setupRecording(self):
        self.stableBumpPop.record({'spikes','v'})
        self.inhibPop.record({'spikes','v'})
        if (self.hasOvercome):
            self.overcomePop.record({'spikes','v'})
        #self.overcomePop.record(['spikes','v','gsyn_exc','gsyn_inh'])

    def getStableBumpPop(self):
        return self.stableBumpPop

